import numpy as np


from pge.model.growth.basic import BUnGrowth, BDiGrowth
from pge.model.growth.basic_fix import FixUnGrowth, FixDiGrowth


class PAGrowth(BUnGrowth):
    def choice(self, graph, sz):
        probs = np.array([graph.count_in_degree(node) for node in graph.get_ids()]) + self.param
        probs = probs / np.sum(probs)
        return np.random.choice(graph.get_ids(), sz, replace=False, p=probs)


class PADiGrowth(BDiGrowth):
    def choice(self, graph, sz, tp="in"):
        if tp == "in":
            probs = np.array([graph.count_in_degree(node) for node in graph.get_ids()]) + self.param[0]
        else:
            probs = np.array([graph.count_out_degree(node) for node in graph.get_ids()]) + self.param[1]
        probs = probs / np.sum(probs)
        return np.random.choice(graph.get_ids(), sz, replace=False, p=probs)


class PAFixUnGrowth(FixUnGrowth):
    def choice(self, graph, sz):
        probs = graph.get_attributes("dg") + self.param
        probs = probs / np.sum(probs)
        return np.random.choice(graph.get_ids(stable=True), sz, replace=False, p=probs)

    def clean(self, graph):
        graph = self.prep(graph)
        if self.param != 0:
            return []

        dels = graph.get_ids(stable=True)[(graph.get_attributes("dg") == 0)]
        return dels

    def prep(self, graph):
        graph.set_attrs(
            "dg", {node: graph.count_in_degree(node) for node in graph.get_ids()}
        )
        return graph


class PAFixDiGrowth(FixDiGrowth):
    def choice(self, graph, sz, tp="in"):
        probs = graph.get_attributes("dg_" + tp) + self.param[tp != "in"]
        probs = probs / np.sum(probs)
        return np.random.choice(graph.get_ids(stable=True), sz, replace=False, p=probs)

    def clean(self, graph):
        if np.sum(self.param) != 0:
            return []

        graph = self.prep(graph)
        dels = graph.get_ids(stable=True)[(graph.get_attributes("dg_in") == 0)
                                         & (graph.get_attributes("dg_out") == 0)]
        return dels

    def prep(self, graph):
        graph.set_attrs(
            "dg_in", {node: graph.count_in_degree(node) for node in graph.get_ids()}
        )
        graph.set_attrs(
            "dg_out", {node: graph.count_out_degree(node) for node in graph.get_ids()}
        )
        return graph



import numpy as np

from pge.wlk.walker import RandomWalk


class RandomNeighbourNode(RandomWalk):
    def generate(self, n, st=None, unique=False):
        if st is None:
            st = np.random.choice(self.gr.get_ids())
        res = [st]

        while True:
            if self.gr.count_out_degree(res[-1]) == 0:
                res.append(res[0])
            else:
                res.append(np.random.choice(self.gr.get_out_degrees(res[-1])))

            if (unique and np.unique(res).size == n) or (~unique and len(res) == n):
                break
        if unique:
            return np.unique(res)
        else:
            return np.array(res)



import networkx as nx
import numpy as np


def information_centrality(graph):
    return nx.information_centrality(graph.get_nx_graph())


def closeness_centrality(graph):
    return nx.closeness_centrality(graph.get_nx_graph())


def closeness2_centrality(graph):
    graph = graph.get_nx_graph()
    path_length = nx.single_source_shortest_path_length
    nodes = graph.nodes

    closeness_centrality2 = {}
    for n in nodes:
        sp = path_length(graph, n)
        totsp = np.sum(np.power(list(sp.values()), 2))/(len(sp) - 1.0)
        closeness_centrality2[n] = totsp
    return closeness_centrality2



import numpy as np
import networkx as nx

from pge.init.classes.graph import SGraph


class CustomCreator:
    @staticmethod
    def partitioned_grow(types, prt):
        graph = nx.Graph()
        count = 0
        for tp in types:
            nw_graph = tp[0](*tp[1])
            nw_graph.set_attrs("prt", count)
            graph = nx.union(graph, nw_graph.get_nx_graph(), rename=(None, tp[2]))
            count += 1

        graph = SGraph(graph)
        nodes = graph.get_ids(stable=True)
        dg = [int(prt * graph.count_out_degree(node)) + 1 for node in nodes]
        dg = np.array([np.random.choice(i, 1, p=(np.arange(i) + 1)[::-1] / (np.sum(np.arange(i) + 1)))[0] for i in dg])
        if np.sum(dg) % 2 != 0:
            dg[0] += 1
        prts = graph.get_attributes("prt")

        for i in np.arange(nodes.size):
            if dg[i] == 0:
                continue

            indx = np.arange(nodes.size)[(prts != prts[i]) & (dg > 0)]
            if indx.size == 0:
                continue
            adds = np.random.choice(indx, dg[i])

            for add in adds:
                if dg[add] > 0:
                    graph.add_edge(nodes[i], nodes[add])
                    dg[add] -= 1
            dg[i] = 0
        return graph

    @staticmethod
    def partitioned_di_grow(types, prt, path, rep=1):
        graph_ = nx.DiGraph()
        count = 0
        for tp in types:
            nw_graph = tp[0](*tp[1])
            nx.write_graphml(nw_graph.get_nx_graph(), path + str(tp[2]) + ".graphml")
            nw_graph.set_attrs("prt", count)
            graph_ = nx.union(graph_, nw_graph.get_nx_graph(), rename=(None, tp[2] + "-"))
            count += 1

        try:
            prt = [prt_ for prt_ in prt]
        except:
            prt = [prt]

        for prt_ in prt:
            for _ in np.arange(rep):
                graph = SGraph(graph_)
                nodes = graph.get_ids(stable=True)
                dg_out = [int(prt_ * graph.count_out_degree(node)) + 1 for node in nodes]
                dg_out = np.array(
                    [np.random.choice(i, 1, p=(np.arange(i) + 1)[::-1] / (np.sum(np.arange(i) + 1)))[0] for i in
                     dg_out])
                dg_in = [int(prt_ * graph.count_in_degree(node)) + 1 for node in nodes]
                dg_in = np.array(
                    [np.random.choice(i, 1, p=(np.arange(i) + 1)[::-1] / (np.sum(np.arange(i) + 1)))[0] for i in dg_in])
                print(prt_, np.sum(dg_in), np.sum(dg_out))

                prts = graph.get_attributes("prt")
                id_ns = np.arange(nodes.size)
                np.random.shuffle(id_ns)
                for i in id_ns:
                    if dg_in[i] != 0:
                        indx = np.arange(nodes.size)[(prts != prts[i]) & (dg_out > 0)]
                        if indx.size == 0:
                            continue

                        adds = np.random.choice(indx, dg_in[i])

                        for add in adds:
                            if dg_out[add] > 0:
                                graph.add_edge(nodes[i], nodes[add])
                                dg_out[add] -= 1
                        dg_in[i] = 0

                nx.write_graphml(graph.get_nx_graph(), path + "custom_" + str(prt_) + "_" + str(_) + ".graphml")



import pandas as pd
import numpy as np
import networkx as nx


class SGraph:
    def __init__(self, gr):
        self.gr = gr
        self.p = None
        self.pi = None

    def clean_copy(self):
        G = self.gr.__class__()
        G.add_edges_from(self.gr.edges())
        return SGraph(G)

    def directed(self):
        return nx.is_directed(self.get_nx_graph())

    def to_undirected(self):
        if self.directed():
            return SGraph(self.gr.to_undirected())
        return SGraph(self.gr.copy())

    def number_of_edges(self):
        return self.gr.number_of_edges()

    def clean(self):
        self.gr.clear()

    def subgraph(self, nodes):
        sub = self.gr.subgraph(nodes).copy()
        return SGraph(sub)

    def get_ids(self, stable=False):
        if stable:
            return np.array(list(self.gr.nodes()))
        else:
            return self.gr.nodes()

    def get_nodes_with(self, attr, value):
        ids = self.get_ids(stable=True)
        return ids[self.get_attributes(attr) == value]

    def get_edges(self, stable=False):
        if stable:
            return np.array(list(self.gr.edges(data=False)))
        else:
            return self.gr.edges(data=False)

    def has_node(self, node):
        return self.gr.has_node(node)

    def set_attr(self, ind, attr, value):
        self.gr.nodes[ind][attr] = convert(value)

    def set_attrs(self, attr, values):
        nx.set_node_attributes(self.gr, name=attr, values=values)

    def get_attr(self, ind, attr):
        if hasattr(ind, "__iter__") and not isinstance(ind, str):
            res = []
            for i in ind:
                try:
                    res.append(self.gr.nodes[i][attr])
                except:
                    res.append(0)
            return np.array(res)

        try:
            return self.gr.nodes[ind][attr]
        except:
            return 0

    def get_node_data(self, ind):
        return self.gr.nodes[ind]

    def add_node(self, id_n):
        self.gr.add_node(id_n)

    def add_nodes(self, ids):
        self.gr.add_nodes_from(ids)

    def del_node(self, id_n):
        self.gr.remove_node(id_n)

    def del_nodes(self, ids):
        self.gr.remove_nodes_from(ids)

    def copy(self):
        return SGraph(self.gr.copy())

    def get_attributes(self, attr, nodes=None, noneval=0):
        if nodes is None:
            nodes = self.gr.nodes()

        res = []
        for x in nodes:
            try:
                res.append(self.gr.nodes[x][attr])
            except:
                res.append(noneval)
        return np.array(res)

    def get_edge_attributes(self, attr, noneval=0):
        res = []
        for x in self.gr.edges(data=False):
            try:
                res.append(self.get_edge_data(x[0], x[1], attr))
            except:
                res.append(noneval)
        return np.array(res)

    def get_attrs_dict(self, attr):
        return nx.get_node_attributes(self.gr, attr)

    def get_nx_graph(self):
        return self.gr

    def get_prob_matrix(self):
        if self.p is None:
            p = nx.to_numpy_matrix(self.gr)
            self.p = np.divide(p, np.sum(p, axis=0))
        return self.p

    def get_stat_dist(self):
        if self.pi is None:
            A = np.subtract(
                np.transpose(self.get_prob_matrix()), np.diag(np.ones(self.size()))
            )
            A = np.vstack([A, np.ones(self.size())])

            b = np.zeros(self.size() + 1)
            b[-1] = 1

            try:
                self.pi = np.linalg.lstsq(A, b, rcond=None)[0]
            except:
                self.pi = np.ones(self.size())
        return self.pi

    def add_edge(self, u, v):
        self.gr.add_edge(u, v)

    def add_edges(self, edges):
        self.gr.add_edges_from(edges)

    def get_edge_data(self, u, v, frm, bsc=0):
        try:
            return self.gr[u][v][frm]
        except:
            return bsc

    def set_edge_data(self, u, v, to, value):
        self.gr[u][v][to] = value

    def del_edge(self, u, v):
        self.gr.remove_edge(u, v)

    def del_edges(self, edges):
        self.gr.remove_edges_from(edges)

    def add_path(self, path):
        self.gr.add_path(path)

    def size(self):
        return self.gr.number_of_nodes()

    def size_edge(self):
        return self.gr.number_of_edges()

    def count_in_degree(self, ind, w=None):
        if w is None:
            return self.get_in_degrees(ind).size
        res = self.get_in_degrees(ind, w=w, bsc=1)
        return np.sum(res)

    def count_out_degree(self, out, w=None,):
        if w is None:
            return self.get_out_degrees(out).size
        res = self.get_out_degrees(out, w=w, bsc=1)
        return np.sum(res)

    def get_count_edges(self):
        return self.gr.number_of_edges()

    def get_in_degrees(self, ind, w=None, un=True, bsc=0):
        if self.directed():
            res = [x[0] for x in self.gr.in_edges(ind)]
        else:
            res = [x[0] for x in self.gr.edges(ind) if x[0] != ind] + [
                x[1] for x in self.gr.edges(ind) if x[1] != ind
            ]

        if un:
            res = np.unique(res)
        if w is None:
            return np.array(res)
        else:
            return np.array([self.get_edge_data(res, ind, w, bsc=bsc)])

    def get_out_degrees(self, out, w=None, un=True, bsc=0):
        if self.directed():
            res = [x[1] for x in self.gr.out_edges(out)]
        else:
            res = [x[0] for x in self.gr.edges(out) if x[0] != out] + [
                x[1] for x in self.gr.edges(out) if x[1] != out
            ]

        if un:
            res = np.unique(res)
        if w is None:
            return np.array(res)
        else:
            return np.array([self.get_edge_data(out, res, w, bsc=bsc)])

    def get_data(self, attrs=None, nodes=None):
        if nodes is None:
            nodes = self.gr.nodes()

        if attrs is None:
            attrs = (self.gr.node[nodes[0]]).keys()

        data = []
        for attr in attrs:
            data.append((attr, []))
        ids = []

        for node in nodes:
            ids.append(node)
            for data_line in data:
                data_line[1].append(self.gr.node[node][data_line[0]])

        df = pd.DataFrame(ids, columns=["id"])
        for data_line in data:
            df[data_line[0]] = pd.Series(data_line[1], index=df.index)
        return df

    def save_attrs(self, path, sep=","):
        data = self.get_data()
        data.to_csv(path, sep=sep, index=False)

    def del_attrs(self, attrs):
        if not hasattr(attrs, "__iter__"):
            attrs = [attrs]

        for n in self.get_ids():
            for attr in attrs:
                del self.gr.nodes[n][attr]

    def get_blk_attrs(self, frm, nodes, typ, part=True):
        xs = np.array([])
        blks = np.array([])

        count = 0
        if typ == "sl" or typ == "sl-ext":
            for j in nodes:
                subs = self.get_in_degrees(j)
                if typ == "sl-ext":
                    subs = np.append(j, subs)
                if part:
                    subs = subs[np.isin(subs, nodes)]
                if subs.size == 0:
                    continue
                res = self.get_attributes(frm, subs)
                blks = np.append(blks, count * np.ones(res.size))
                xs = np.append(xs, res)
                count = count + 1

        if typ == "b-sl" or typ == "b-sl-ext":
            for j in nodes:
                subs = self.get_out_degrees(j)
                if typ == "b-sl-ext":
                    subs = np.append(j, subs)
                if part:
                    subs = subs[np.isin(subs, nodes)]
                if subs.size == 0:
                    continue
                res = self.get_attributes(frm, subs)
                blks = np.append(blks, count * np.ones(res.size))
                xs = np.append(xs, res)
                count = count + 1

        if typ == "dj":
            subs = []
            blk = []
            subs_ = np.array([])
            blk_ = []

            for j in nodes:
                sub = self.get_out_degrees(j)

                if sub.size > 1:
                    nad = True
                    for prt in np.arange(len(subs)):
                        if np.sum(np.isin(sub, subs[prt])) > 1:
                            nad = False
                            subs[prt] = np.unique(np.append(subs[prt], sub))
                            blks = np.append(blks, blk[prt])
                            break
                    if nad:
                        subs.append(sub)
                        blk.append(count)
                        blks = np.append(blks, count)
                        count += 1
                else:
                    nex = True
                    for prt in np.arange(subs_.size):
                        if subs_[prt] == sub[0]:
                            nex = False
                            blks = np.append(blks, blk_[prt])
                            break

                    if nex:
                        blks = np.append(blks, count)
                        subs_ = np.append(subs_, sub)
                        blk_.append(count)
                        count += 1
                xs = np.append(xs, self.get_attr(j, frm))
        return xs, blks

    def get_ln_attrs(self, frm, nodes, count=False):
        xs = []
        lv = np.array([])

        for node in nodes:
            xs.append(self.get_attributes(frm, node))
            lv = np.append(lv, self.get_attributes(frm, node))
        lv = np.unique(lv, return_counts=count)
        return np.array(xs), lv

    def get_all_short_pathes(self, nodes, plain=True):
        if len(nodes) < 2:
            return [], []

        pathes, ln = [], []

        for node in nodes:
            pathes_, ln_ = self.get_short_paths(node, nodes)
            pathes += pathes_
            if plain:
                ln += ln_
            else:
                if len(ln_) != 0:
                    ln += [ln_]
        return pathes, np.array(ln)

    def get_short_line_paths(self, root, nodes=None):
        if nodes is None:
            nodes = self.get_ids(stable=True)

        pathes, ln = [], []
        for node in nodes:
            if root == node:
                continue
            path = next(
                nx.all_shortest_paths(
                    self.gr, source=root, target=node, method="bellman-ford"
                )
            )
            if isinstance(path[0], list):
                for path_ in path:
                    pathes.append(np.array(path_))
                    ln.append(len(path_) - 1)
            else:
                pathes.append(np.array(path))
                ln.append(len(path) - 1)
        return pathes, ln

    def get_short_paths(self, root, nodes):
        if len(nodes) < 2:
            return [], []

        pathes, ln = [], []
        done = [root]
        step = [[root]]

        while len(step) != 0:
            if len(pathes) == len(nodes) - 1:
                break

            step_ = []
            for st in step:
                for res in self.get_in_degrees(st[-1]):
                    if res not in done:
                        done.append(res)
                        st_ = st.copy()
                        st_.append(res)

                        if res in nodes:
                            pathes.append(st_)
                            ln.append(len(st_) - 1)
                        else:
                            step_.append(st_)
            step = step_
        return pathes, ln

    def get_short_paths2(self, root, nodes):
        if len(nodes) < 2:
            return [], []

        pathes, ln = [], []
        done = [root]
        step = [[root]]

        while len(step) != 0:
            if len(pathes) == len(nodes) - 1:
                break

            step_ = []
            for st in step:
                for res in self.get_in_degrees(st[-1]):
                    if res not in done:
                        done.append(res)
                        st_ = st.copy()
                        st_.append(res)

                        if res in nodes:
                            pathes.append(st_)
                            ln.append(len(st_) - 1)
                        else:
                            step_.append(st_)
            step = step_
        return pathes, ln


def convert(value):
    try:
        return float(value)
    except:
        return str(value)



import numpy as np

from dtg.stationary.dierckx import stat_exponential
from dtg.stationary.giratis import stat
from pge.cluster.cycle import all_cycles_undirected
from pge.ranks.extrem_onl import NodeExInfo


class CircleEx(NodeExInfo):
    @staticmethod
    def get_exes_comm(gr, nodes, params):
        res = []
        cycles = all_cycles_undirected(gr.subgraph(nodes))

        for params_ in params:
            exes = []
            alps = []

            for cycle in cycles:
                cls = gr.get_attributes(params_[0], cycle)
                for _ in np.arange(8):
                    cls = np.append(cls, cls)

                cls = [gr.get_attributes(params_[0], cycle) for cycle in cycles]
                exes.append(params_[1].estimate(cls))
                alps.append(params_[2].estimate(cls))

            exes = np.array(exes)
            alps = np.array(alps)

            res.append(np.mean(exes[alps == np.max(alps)]))
        return res

    @staticmethod
    def get_view(gr, nodes, params):
        res = []

        # ds = np.array([gr.count_in_degree(node) for node in gr.get_ids()])
        ds = np.array([gr.count_in_degree(node) for node in nodes])
        for params_ in params:
            cls = gr.get_attributes(params_[0], nodes)

            for u in np.unique(cls)[::-1]:
                ts = np.array(gr.get_all_short_pathes(nodes[cls > u])[1])

                res.append(np.mean(ds[cls > u]) / ts.size)
        return res, np.unique(cls)[::-1]

    @staticmethod
    def get_test_comm(gr, nodes, level, param):
        if nodes.size == 0:
            return (0, 0), (0, 0)

        cls = gr.get_attributes(param, nodes)
        ts = gr.get_all_short_pathes(nodes[cls > level], plain=True)[1]
        if ts.size == 0:
            return (0, 0), (stat(cls), np.mean(cls))

        q = np.sum(gr.get_attributes(param) > level) / gr.size()
        return (
            (
                stat_exponential(ts, q),
                q
                * np.sum([np.sum(ti) for ti in ts])
                / np.sum([np.size(ti) for ti in ts]),
            ),
            (stat(cls), np.mean(cls)),
        )



import itertools

import networkx as nx
import numpy as np


def all_cycles_undirected(graph):
    basic_cycles = nx.minimum_cycle_basis(graph.get_nx_graph())

    cycles = []
    for bas in basic_cycles:
        cycles.append(bas)

    for i in np.arange(2, len(basic_cycles) + 1):
        for sums_cycles in itertools.combinations(basic_cycles, i):
            G_sub = nx.Graph()

            for cycle in sums_cycles:
                ln = len(cycle)
                for j in np.arange(ln)[::-1]:
                    if G_sub.has_edge(cycle[j], cycle[j - 1]):
                        G_sub[cycle[j]][cycle[j - 1]].update(
                            {"count": G_sub[cycle[j]][cycle[j - 1]]["count"] + 1}
                        )
                    else:
                        G_sub.add_edge(cycle[i], cycle[i - 1], count=1)

            if nx.is_connected(G_sub):
                dls = [dt for dt in G_sub.edges.data() if dt[2]["count"] > 1]
                if len(dls) > 0:
                    for dl in dls:
                        G_sub.remove_edge(dl[0], dl[1])

                    add_cycles = nx.minimum_cycle_basis(G_sub)
                    if len(add_cycles) > 1:
                        continue

                    cycles.append(add_cycles[0])

    return cycles



import networkx as nx
import numpy as np

from pge.init.classes.graph import SGraph


class Status(object):
    def __init__(self):
        self.node2com = dict([])
        self.total_weight = 0
        self.out_degrees = dict([])
        self.in_degrees = dict([])
        self.in_gdegrees = dict([])
        self.out_gdegrees = dict([])
        self.internals = dict([])
        self.loops = dict([])

    def copy(self):
        new_status = Status()
        new_status.node2com = self.node2com.copy()
        new_status.internals = self.internals.copy()
        new_status.in_degrees = self.in_degrees.copy()
        new_status.out_degrees = self.out_degrees.copy()
        new_status.in_gdegrees = self.in_gdegrees.copy()
        new_status.out_gdegrees = self.out_gdegrees.copy()
        new_status.total_weight = self.total_weight
        new_status.loops = self.loops

    def init(self, graph):
        count = 0
        self.total_weight = graph.size_edge()

        for node in graph.get_ids():
            self.node2com[node] = count
            out_deg = graph.count_out_degree(node, w="weight")
            in_deg = graph.count_in_degree(node, w="weight")
            self.out_degrees[count] = out_deg
            self.in_degrees[count] = in_deg
            self.in_gdegrees[node] = in_deg
            self.out_gdegrees[node] = in_deg

            self.loops[node] = graph.get_edge_data(node, node, "weight", bsc=0)
            self.internals[count] = self.loops[node]
            count += 1


__PASS_MAX = -1
__MIN = 0.0000001


def partition_at_level(dendrogram, level):
    partition = dendrogram[0].copy()
    for index in range(1, level + 1):
        for node, community in partition.items():
            partition[node] = dendrogram[index][community]
    return partition


def modularity(partition, graph):
    inc = dict([])
    deg = dict([])

    links = graph.size_edge()
    if links == 0:
        raise ValueError("A graph without link has an undefined modularity")

    for node in graph.get_ids():
        com = partition[node]
        node_in_deg = graph.count_in_degree(node)
        for neighbor in graph.get_out_degrees(node):
            if partition[neighbor] == com:
                deg[com] = deg.get(com, 0.) + node_in_deg * graph.count_out_degree(neighbor)
                inc[com] = inc.get(com, 0.) + 1

    res = 0.
    for com in set(partition.values()):
        res += (inc.get(com, 0.) / links) - (deg.get(com, 0.) / (links ** 2))
    return res


def best_partition_to(graph, starts=10, com_sz=1):
    part = {node: i for i, node in enumerate(graph.get_ids(stable=True))}
    mod = modularity(part, graph)

    for _ in range(starts):
        dendo = generate_dendrogram(graph)
        part_ = partition_at_level(dendo, len(dendo)-1)
        ln = len(set(part_.values()))
        if ln >= com_sz:
            mod_ = modularity(part_, graph)
            if mod_ > mod:
                mod = mod_
                part = part_
    return part, mod


def generate_dendrogram(graph):
    if graph.number_of_edges() == 0:
        part = dict([])
        for i, node in enumerate(graph.get_ids(stable=True)):
            part[node] = i
        return [part]

    current_graph = graph.clean_copy()
    status = Status()
    status.init(current_graph)
    status_list = list()
    __one_level(current_graph, status)

    new_mod = __modularity(status)
    partition = __renumber(status.node2com)
    status_list.append(partition)
    mod = new_mod
    current_graph = induced_graph(partition, current_graph)
    status.init(current_graph)

    while True:
        __one_level(current_graph, status)
        new_mod = __modularity(status)
        if new_mod - mod < __MIN:
            break
        partition = __renumber(status.node2com)
        status_list.append(partition)
        mod = new_mod
        current_graph = induced_graph(partition, current_graph)
        status.init(current_graph)
    return status_list


def induced_graph(partition, graph):
    ret = nx.DiGraph()
    ret.add_nodes_from(partition.values())

    for node1, node2 in graph.get_nx_graph().edges():
        wl = graph.get_edge_data(node1, node2, "weight", bsc=1)
        com1 = partition[node1]
        com2 = partition[node2]
        w_prec = ret.get_edge_data(com1, com2, {"weight": 0}).get("weight", 1)
        ret.add_edge(com1, com2, **{"weight": w_prec + wl})
    return SGraph(ret)


def __renumber(dictionary):
    values = set(dictionary.values())
    target = set(range(len(values)))

    if values == target:
        ret = dictionary.copy()
    else:
        renumbering = dict(zip(target.intersection(values), target.intersection(values)))
        renumbering.update(dict(zip(values.difference(target), target.difference(values))))
        ret = {k: renumbering[v] for k, v in dictionary.items()}
    return ret


def __one_level(graph, status):
    modified = True
    nb_pass_done = 0
    cur_mod = __modularity(status)
    new_mod = cur_mod

    while modified and nb_pass_done != __PASS_MAX:
        cur_mod = new_mod
        modified = False
        nb_pass_done += 1

        ids = graph.get_ids(stable=True)
        np.random.shuffle(ids)
        for node in ids:
            com_node = status.node2com[node]
            in_degc_totw = status.in_gdegrees.get(node, 0.) / status.total_weight
            out_degc_totw = status.out_gdegrees.get(node, 0.) / status.total_weight
            neigh_communities = __neighcom(node, graph, status)

            remove_cost = -  neigh_communities.get(com_node, 0) + (
                        status.out_degrees.get(com_node, 0.) - status.out_gdegrees.get(node, 0.)) * in_degc_totw + (
                        status.in_degrees.get(com_node, 0.) - status.in_gdegrees.get(node, 0.)) * out_degc_totw
            __remove(node, com_node, neigh_communities.get(com_node, 0.), status)

            best_com = com_node
            best_increase = 0

            for com, dnc in __randomize(neigh_communities.items()):
                incr = remove_cost + dnc - (status.out_degrees.get(com, 0.) * in_degc_totw + status.in_degrees.get(com, 0.) * out_degc_totw)
                if incr > best_increase:
                    best_increase = incr
                    best_com = com
            __insert(node, best_com, neigh_communities.get(best_com, 0.), status)
            if best_com != com_node:
                modified = True
        new_mod = __modularity(status)
        if new_mod - cur_mod < __MIN:
            break


def __neighcom(node, graph, status):
    weights = {}
    for neighbor in graph.get_out_degrees(node):
        if neighbor != node:
            edge_weight = graph.get_edge_data(node, neighbor, "weight", bsc=1)
            neighborcom = status.node2com[neighbor]
            weights[neighborcom] = weights.get(neighborcom, 0) + edge_weight
    return weights


def __remove(node, com, weight, status):
    status.in_degrees[com] = status.in_degrees.get(com, 0.) - status.in_gdegrees.get(node, 0.)
    status.out_degrees[com] = status.out_degrees.get(com, 0.) - status.out_gdegrees.get(node, 0.)
    status.internals[com] = status.internals.get(com, 0.) - weight - status.loops.get(node, 0.)
    status.node2com[node] = -1


def __insert(node, com, weight, status):
    status.node2com[node] = com
    status.in_degrees[com] = status.in_degrees.get(com, 0.) + status.in_gdegrees.get(node, 0.)
    status.out_degrees[com] = status.out_degrees.get(com, 0.) + status.out_gdegrees.get(node, 0.)
    status.internals[com] = status.internals.get(com, 0.) + weight + status.loops.get(node, 0.)


def __modularity(status):
    links = status.total_weight
    result = 0.
    if links > 0:
        for community in set(status.node2com.values()):
            in_degree = status.internals.get(community, 0.)
            degree = status.in_degrees.get(community, 0.) * status.out_degrees.get(community, 0.)
            result += (in_degree / links) - (degree / (links ** 2))
    return result


def __randomize(items):
    randomized_items = list(items)
    np.random.shuffle(randomized_items)
    return randomized_items



import numpy as np
import networkx as nx

from dtg.tail.estimate.hill import HillEstimator
from dtg.tail.mse import boot_estimate
from pge.init.classes.graph import SGraph
from pge.init.add import actual


def powerlaw(n, alp, st=1, ac=0.3, status=None, status_back=False):
    if status is None:
        alp_ = alp + 0.5
    else:
        alp_ = status
    while True:
        rs = np.trunc(nx.utils.powerlaw_sequence(n, alp_)) - 1 + st
        alp2 = boot_estimate(
            HillEstimator,
            rs,
            1 / 2,
            2 / 3,
            50,
            speed=False
        )[0]
        if np.abs(alp2-alp) <= ac:
            break
        else:
            if alp2 > alp:
                alp_ -= 0.15
            else:
                alp_ += 0.15
    if status_back:
        return rs, alp_
    return rs


def powerlaw_degree(n, alpha, beta, ac=0.3, status=(None, None), rep=10):
    out_d, out_status = powerlaw(n, beta, 1, ac=ac, status_back=True, status=status[0])
    ind_d, in_status = powerlaw(n, alpha, 1, ac=ac, status_back=True, status=status[1])

    zv = np.sum(np.subtract(out_d, ind_d))
    for _ in np.arange(rep):
        out_d_, out_status = powerlaw(n, beta, int((beta > alpha))*(np.abs(zv)//n) + 1, ac=ac,
                                      status_back=True, status=out_status)
        if np.abs(np.sum(np.subtract(out_d_, ind_d))) < np.abs(zv):
            out_d = out_d_
            zv = np.sum(np.subtract(out_d, ind_d))
        ind_d_, in_status = powerlaw(n, alpha, int((beta < alpha))*(np.abs(zv)//n) + 1, ac=ac,
                                     status_back=True, status=in_status)
        if np.abs(np.sum(np.subtract(out_d, ind_d_))) < np.abs(zv):
            ind_d = ind_d_
            zv = np.sum(np.subtract(out_d, ind_d))

    print(np.sum(np.subtract(out_d, ind_d)))

    return ind_d, out_d, (in_status, out_status)


class Creator:
    @staticmethod
    def tbt(nm, alpha, beta, prm=0.7):
        k = 1 - min([0.5, 1 - alpha ** -1, 1 - beta ** -1]) / 2
        zv = nm

        status = (None, None)
        while np.abs(zv) > nm ** (1 - k + prm * k):
            ind_d, out_d, status = powerlaw_degree(nm, alpha, beta, status=status)
            zv = np.sum(np.subtract(ind_d, out_d))

        i_s = np.random.choice(nm, min(int(np.abs(zv)), nm), replace=False)
        for i in i_s:
            if zv < 0:
                ind_d[i] = ind_d[i] + 1
            else:
                out_d[i] = out_d[i] + 1

        gr = SGraph(nx.DiGraph())
        gr.add_nodes(np.arange(nm))

        for i in np.arange(nm):
            if ind_d[i] == 0:
                continue
            i_s = np.setdiff1d(np.arange(nm), i)
            np.random.shuffle(i_s)

            for j in i_s:
                if out_d[int(j)] > gr.count_out_degree(j):
                    gr.add_edge(j, i)
                if ind_d[int(j)] > gr.count_in_degree(j):
                    gr.add_edge(i, j)

        return gr

    @staticmethod
    def make_from_data(network, pre="../", nm=None, sigma=None):
        path = pre + "pge/samples/networks/web-"
        if network == "BS":
            path += "BerkStan.txt"
        elif network == "S":
            path += "Stanford.txt"
        elif network == "G":
            path += "Google.txt"
        else:
            path = network
        gr = nx.read_edgelist(path, create_using=nx.DiGraph())
        print("loaded")

        if nm is None:
            res = SGraph(gr)
        else:
            choiced = np.random.choice(gr.nodes())
            res = Creator.create_subgraph(gr, choiced, nm)

        actual(res, sigma)
        print("attrs added")
        return res

    @staticmethod
    def simple_graph(network, pre="../", typ=nx.DiGraph, dl=" ", clean=False):
        path = pre + "pge/samples/networks/web-"
        if network == "BS":
            path += "BerkStan.txt"
        elif network == "S":
            path += "Stanford.txt"
        elif network == "G":
            path += "Google.txt"
        else:
            path = network
        res = nx.read_edgelist(path, create_using=typ(), delimiter=dl)

        sub = max(nx.connected_components(res), key=len)
        res = SGraph(res).subgraph(sub)

        if clean:
            res.del_attrs(["pos"])
        return res

    @staticmethod
    def fire_forest(p_f, p_r, n):
        graph = nx.DiGraph()
        graph.add_nodes_from(np.arange(n))
        graph = SGraph(graph)

        for i in np.arange(1, n):
            add_n = np.array([])
            if i - 1 == 0:
                ws = np.array([0])
            else:
                ws = []  # ws = np.array([rand_int(mn=0, mx=i - 1, sz=1)])

            while ws.size > 0:
                n_ws = np.array([])
                for w in ws:
                    x = graph.get_in_degrees(w)
                    sub_x = np.random.geometric(p_f) - 1
                    if sub_x < x.size:
                        x = np.random.choice(x, sub_x, replace=False)

                    if x.size != 0:
                        n_ws = np.append(n_ws, x)

                    y = graph.get_out_degrees(w)
                    sub_y = np.random.geometric(p_r) - 1
                    if sub_y < y.size:
                        y = np.random.choice(y, sub_y, replace=False)

                    if y.size != 0:
                        n_ws = np.append(n_ws, y)

                if n_ws.size != 0:
                    n_ws = np.setdiff1d(n_ws, add_n)

                for nl in ws:
                    graph.add_edge(i, int(nl))
                add_n = np.append(add_n, ws)
                ws = n_ws
        return graph

    @staticmethod
    def gnp_random_graph(n, p, directed=True):
        return SGraph(nx.generators.fast_gnp_random_graph(n, p, directed=directed))

    @staticmethod
    def geometric(n, dis, clean=True, dim=2):
        res = nx.random_geometric_graph(n, dis, dim=dim)
        sub = max(nx.connected_components(res), key=len)
        res = SGraph(res).subgraph(sub)

        if clean:
            res.del_attrs(["pos"])
        return res

    @staticmethod
    def waxman(n, beta=0.11, alpha=0.1, L=None, clean=True):
        res = nx.waxman_graph(n, beta=beta, alpha=alpha, L=L)
        sub = max(nx.connected_components(res), key=len)
        res = SGraph(res).subgraph(sub)

        if clean:
            res.del_attrs(["pos"])
        return res

    @staticmethod
    def chunglu(seq):
        res = nx.expected_degree_graph(seq, selfloops=False)
        nodes = sorted(nx.connected_components(res), key=len, reverse=True)[0]
        return SGraph(res).subgraph(nodes)

    @staticmethod
    def power_law(n, sigma):
        inds = powerlaw(n, sigma)
        if np.sum(inds) % 2 == 1:
            inds = np.append(inds, [1])

        res = nx.expected_degree_graph([int(u) for u in inds])
        nodes = sorted(nx.connected_components(res), key=len, reverse=True)[0]
        res.remove_edges_from(nx.selfloop_edges(res))
        return SGraph(res).subgraph(nodes)

    @staticmethod
    def load(path, typ=nx.Graph):
        return SGraph(typ(nx.read_graphml(path)))

    @staticmethod
    def load_adj(path, typ=nx.Graph):
        return SGraph(typ(nx.read_adjlist(path)))



import numpy as np

from pge.ranks.rank import estimate_rank


def pgr(graph, c, nodes, params):
    other = graph.get_attr(nodes[0], params[0])
    if graph.directed():
        t = graph.count_out_degree(nodes[0])
    else:
        t = graph.count_degree(nodes[0])
    t = np.where(t == 0, 0.00000001, t)
    ns = c * params[1] / t
    return np.multiply(other, ns*np.ones(other.size))/graph.get_attr(nodes[1], params[0])


def bpgr(graph, c, nodes, params):
    t = graph.count_out_degree(nodes[1])
    if t == 0:
        return np.array([])
    other = 1/graph.get_attr(nodes[0], params[0])
    return other*c*params[1]*graph.get_attr(nodes[1], params[0])/t


def wpgr(graph, c, nodes, params):
    other = graph.get_attr(nodes[0], params[0])
    t = np.multiply(graph.count_out_degree(nodes[0], w=params[0]+"w"), [graph.get_edge_data(i, nodes[1], params[0]+"w") for i in nodes[0]])

    t = np.where(t == 0, 0.00000001, t)
    ns = np.multiply(c * params[1] / t, [graph.get_edge_data(i, nodes[1], params[0]+"w") for i in nodes[0]])
    ns = np.where(ns == 0, 100000000, ns)
    return other, ns


def sumr(graph, c, nodes, params):
    other = np.array([graph.get_edge_data(i, nodes[1], params[0]) for i in nodes[0]])
    return other, np.ones(other.size)*params[1]/graph.get_attr(nodes[1], params[0])


def full(graph, c, nodes, params):
    res = []
    for node in nodes[0]:
        dels = graph.copy()
        dels.del_node(node)

        estimate_rank(dels, tp=params[0], c=c)

        ori = graph.get_attr(nodes[2], params[0])
        nw = graph.get_attr(nodes[2], params[0])
        res.append((ori-nw)/ori)

    return np.array(res), np.ones(len(res))



import numpy as np

from mcg.extreme.estimator.nonopt.simpleblocks import SBlocksEstimate
from mcg.extreme.find.mse import boot
from pge.ranks.extrem_onl import NodeExInfo
from pge.ranks.rank import estimate_rank


class GravNodeEx(NodeExInfo):
    @staticmethod
    def get_ex(gr, root, frm, c, stab=30, entity=SBlocksEstimate,
               full=200, frm_stub=True, view=5, part=True, typ="sl"):

        exes = np.array([])
        ids = np.array([])
        nxt = [np.array([]), np.array([])]
        rank = gr.get_attr(root, frm)

        if rank == 0:
            return np.array([]), np.array([]), np.array([]), np.array([])

        chosen = [root, 0]
        bls = np.array([])
        clust = np.array([])

        count = 0
        count_ = 0
        while chosen is not None:
            ids = np.append(ids, chosen[0])
            k = gr.get_in_degrees(chosen[0])

            k = np.setdiff1d(k, ids)
            k = np.setdiff1d(k, nxt[0])

            if k.size > 0:
                subs = gr.copy()
                subs.delete_node(chosen[0])
                estimate_rank(subs, tp=frm, c=c)
                other = (rank - subs.get_attr(root, frm))/rank

                nxt[0] = np.append(nxt[0], k)
                nxt[1] = np.append(nxt[1], other)

            if nxt[0].size == 0:
                chosen = None
            else:
                ind = np.argmax(nxt[1])
                chosen = [nxt[0][ind], nxt[1][ind]]
                nxt[0] = np.delete(nxt[0], ind)
                nxt[1] = np.delete(nxt[1], ind)

            if frm_stub and ids.size <= stab and chosen is not None:
                continue

            ranks, blks = gr.get_blk_attrs(frm, ids, typ)

            clust = np.append(clust, entity.get_avg_blk(blks))
            us = np.unique(ranks)
            res = boot(entity, (ranks, blks), 1 / 3, 2 / 3, 100, us)
            mses = np.array([res[i][0] for i in np.arange(us.size)])
            ex = np.array([res[i][1] for i in np.arange(us.size)])

            i = np.argmin(mses)
            exes = np.append(exes, ex[i])
            bls = np.append(bls, entity.back((ranks, blks), us[i]))

            if ids.size > full:
                break

            if np.abs((exes[-1] - 1 / clust[-1])) < 0.01 or np.abs(1 - exes[-1] - 1 / clust[-1]) < 0.01:
                count += 1
            else:
                if count != 0:
                    count_ += 1

                count = 0

            if count_ == view:
                exes = exes[:exes.size - 1]
                clust = clust[:clust.size - 1]
                bls = bls[:bls.size - 1]
                ids = ids[:ids.size - 1]
                break

        return ids, exes, bls, clust



import numpy as np

from dtg.dependence.acf import acf


def autocorrelation(graph, attr, h=None):
    nodes = graph.get_ids(stable=True)
    mxes = []

    comms = set([(node,) for node in nodes])
    count = 0
    while count < graph.size():
        print(len(mxes))
        new_comms = set()

        mx = np.max([np.max(graph.get_attributes(attr, comm)) for comm in comms])
        for comm in comms:
            if len(mxes) < count + 1:
                mxes.append(mx)
                new_comms.add(comm)
            else:
                if mxes[count] == mx:
                    new_comms.add(comm)
                if mxes[count] > mx:
                    new_comms = set(comm)
                    mxes[count] = mx

        comms = set()
        for new_comm in new_comms:
            for node_out in graph.get_out_degrees(new_comm, un=True):
                if node_out not in new_comm:
                    comms.add(tuple(np.sort(list(new_comm) + [node_out])))
        count += 1

    return acf(np.array(mxes), h)



import numpy as np

from pge.cluster.connect import phi


class SpreadingModel(object):
    def __init__(self, graph):
        self.graph = graph
        self.status = {}
        self.initial_status = {}
        self.check = 0
        self.ids = self.graph.get_ids(stable=True)
        self.times = []
        self.received = []
        self.messes = []

    def set_initial_status(self, configuration):
        self.initial_status = configuration
        self.check = np.unique([j for i in configuration.values() for j in i]).size

    def init(self):
        self.status = {
            k: self.initial_status[k].copy() for k in self.initial_status.keys()
        }
        self.times = np.zeros(self.ids.size)
        self.received = np.array([len(self.initial_status[node]) for node in self.ids])
        self.messes = self.received.copy()

    def iteration_bunch(self, num_iter, mx=None, prnt=False):
        times = []

        for _ in np.arange(num_iter):
            n = 0
            self.init()

            while np.sum(self.received < self.check) > 0:
                if prnt:
                    print("!", n)

                self.times[self.messes < self.ids.size] += 1
                n += 1
                self.iteration()

                if mx is not None:
                    if n == mx:
                        break

            mx = np.max(self.times)
            times = np.append(times, self.times)
        return times.reshape((num_iter, self.graph.size()))

    def iteration_bunch_comm(self, num_iter, tick, prnt=False):
        ticks = []
        phis = []
        nodes = []

        for _ in np.arange(num_iter):
            n = 0
            self.init()

            while np.sum(self.received < self.check) > 0:
                if prnt:
                    print("!", n)

                if n > tick:
                    break

                n += 1
                self.iteration()

            ticks = np.append(ticks, np.sum(self.received != self.check))
            phis = np.append(
                phis, phi(self.graph, self.ids[self.received != self.check])
            )
            nodes.append(self.ids[self.received != self.check])
        return ticks, phis, nodes

    def iteration_bunch_complex(self, nodes):
        self.init()
        n = 0
        k = 0
        checked = True

        while True:
            self.iteration()
            n += 1

            if nodes is not None:
                if self.finish(nodes) and checked:
                    k = n
                    checked = False
            if self.finish():
                break
        return k, n

    def iteration_timer(self, size=None):
        ns = []
        self.init()

        while np.sum(self.received < self.check) > 0:
            self.iteration()
            ns.append(self.received == self.check)

            if size is not None:
                if len(ns) == size - 1:
                    break
        return ns

    def finish(self, nodes=None):
        if nodes is None:
            nodes = self.graph.get_ids()

        for node in nodes:
            if len(self.status[node]) < self.check:
                return False
        return True

    def iteration(self):
        return None, True



import numpy as np
from scipy.stats import ks_2samp
import networkx as nx

from pge.init.classes.graph import SGraph
from pge.wlk.rnn import RandomNeighbourNode


def check(graph, attr, nm=20, rw=RandomNeighbourNode, iter=100, test=ks_2samp):
    step = int(graph.size()/nm)
    szes = np.arange(step, graph.size(), step=step)
    walker = rw(graph)
    smpls = graph.get_attributes(attr)

    res1, res2 = [], []
    for sz in szes:
        for _ in np.arange(iter):
            sub = graph.get_attributes(attr, walker.generate(sz, unique=True))
            res2.append(test(smpls, sub)[1])
            res1.append(sz)
    return np.array(res1), np.array(res2)


def check_mean(graph, attr, nm=20, rw=RandomNeighbourNode, iter=100, test=ks_2samp):
    res1, res2 = check(graph, attr, nm=nm, rw=rw, iter=iter, test=test)
    un = np.unique(res1)
    nw1, nw2, nw3 = [], [], []
    for u in un:
        nw1.append(np.max(res2[res1 == u]))
        nw2.append(np.mean(res2[res1 == u]))
        nw3.append(np.min(res2[res1 == u]))
    return nw1, nw2, nw3, un


def check2(graph, attr, test=ks_2samp):
    ids = graph.get_ids(stable=True)
    smpls = graph.get_attributes(attr)

    res1, res2 = [], []
    for r in np.arange(1, nx.diameter(graph.get_nx_graph())):
        for root in ids:
            #root = np.random.choice(ids)
            sub = SGraph(nx.ego_graph(graph.get_nx_graph(), root, r))
            sub = graph.get_attributes(attr, sub.get_ids())
            res2.append(test(smpls, sub)[1])
            res1.append(r)
    return np.array(res2), np.array(res1)


def check2_mean(graph, attr, test=ks_2samp):
    res1, res2 = check2(graph, attr, test=test)

    un = np.unique(res2)
    nw1, nw2, nw3 = [], [], []
    for u in un:
        nw1.append(np.max(res1[res2 == u]))
        nw2.append(np.mean(res1[res2 == u]))
        nw3.append(np.min(res1[res2 == u]))
    return nw1, nw2, nw3, un



import numpy as np

from mcg.extreme.estimator.nonopt.blocks import BlocksEstimate
from pge.ranks.extrem_onl import NodeExInfo
from pge.ranks.onlcore.criterion.rcriterion import pgr


def back(graph, chosen, wr):
    return graph.get_out_degrees(chosen, w=wr)


def str_(graph, chosen, wr):
    if graph.directed():
        return graph.get_in_degrees(chosen, w=wr)
    if hasattr(chosen, "__iter__"):
        res = []
        for chosen_ in chosen:
            res = np.append(res, graph.get_degrees(chosen_, wr))
        return np.unique(res)
    return graph.get_degrees(chosen, wr)


class DepNodeEx(NodeExInfo):
    @staticmethod
    def get_ex(gr, root, frm, c=0.85, stab=30, entity=BlocksEstimate,
               full=None, frm_stub=True, view=5, part=True, criterion=pgr, typ="sl", w=False, way=str_):
        if full is None:
            full = gr.size()

        if w:
            wr = frm+"w"
        else:
            wr = None

        exes = np.array([])
        ids = np.array([])
        nxt = [np.array([]), np.array([])]
        rank = gr.get_attr(root, frm)

        if rank == 0:
            return np.array([]), np.array([]), np.array([]), np.array([])

        choosen = [root, 1]
        bls = np.array([])
        clust = np.array([])

        count = 0
        count_ = 0
        while choosen is not None:
            ids = np.append(ids, choosen[0])
            k = way(gr, choosen[0], wr)
            if part:
                k = np.setdiff1d(k, ids)
                k = np.setdiff1d(k, nxt[0])

            if k.size > 0:
                other = criterion(gr, c, (k, choosen[0], root), (frm, choosen[1]))

                nxt[0] = np.append(nxt[0], k)
                nxt[1] = np.append(nxt[1], other)

            if nxt[0].size == 0:
                choosen = None
            else:
                ind = np.argmax(nxt[1])
                choosen = [nxt[0][ind], nxt[1][ind]]
                nxt[0] = np.delete(nxt[0], ind)
                nxt[1] = np.delete(nxt[1], ind)

            if frm_stub and ids.size <= stab and choosen is not None:
                continue

            ranks, blks = gr.get_blk_attrs(frm, ids, typ)
            clust = np.append(clust, entity.get_avg_blk(blks))

            if clust[-1] == 1:
                exes = np.append(exes, 1)
                bls = np.append(bls, 1)
            else:
                ex, u = entity.mse((ranks, blks))
                exes = np.append(exes, ex)
                bls = np.append(bls, entity.back((ranks, blks), u))

            if ids.size >= full:
                break

            if np.abs((exes[-1] - 1 / clust[-1])) < 0.01 or np.abs(1 - exes[-1] - 1 / clust[-1]) < 0.01:
                count += 1
            else:
                if count != 0:
                    count_ += 1
                count = 0

            if count_ == view:
                exes = exes[:exes.size-1]
                clust = clust[:clust.size-1]
                bls = bls[:bls.size-1]
                ids = ids[:ids.size-1]
                break

        return ids, exes, bls, clust



import networkx as nx
import numpy as np
import pandas as pd

from dtg.tail.estimate.hill import HillEstimator
from dtg.tail.mse import boot_estimate
from pge.init.creator import powerlaw
from pge.model.growth.pa import PADiGrowth
from pge.model.growth.pa_com_leader import PAСomDiGrowth
from pge.ranks.rank import estimate_rank


class PAFixComDiGrowth(PAСomDiGrowth):
    def __init__(self, graph, schema, deg, model_param):
        super().__init__(graph, schema, deg, model_param)
        self.chosen = None
        self.frm = "prt"
        self.coms = []

    def prep(self, graph):
        estimate_rank(graph, "one", pers=None)

        self.cur_alps = []
        sets = list(set(graph.get_attributes(self.frm)))
        for com in sets:
            ones = graph.get_attributes("one", graph.get_nodes_with(self.frm, com))
            self.cur_alps.append(
                boot_estimate(
                    HillEstimator,
                    ones,
                    1 / 2,
                    2 / 3,
                    30,
                    speed=False,
                )[0]
            )
            if self.cur_alps[-1] is None:
                self.cur_alps[-1] = np.infty
            else:
                self.cur_alps[-1] = 1 / self.cur_alps[-1]
        self.alps.append(self.cur_alps[np.argmin(self.cur_alps)])
        self.chosen = sets[np.argmin(self.cur_alps)]
        self.coms.append(self.chosen)

        graph.set_attrs(
            "dg_in", {node: graph.count_in_degree(node) for node in graph.get_ids()}
        )
        graph.set_attrs(
            "dg_out", {node: graph.count_out_degree(node) for node in graph.get_ids()}
        )
        return graph

    def save(self, gr, to):
        nx.write_graphml(gr, to + ".graphml")
        prd = pd.DataFrame({"alp": self.alps, "com": self.coms})
        prd.to_csv(to + ".csv")

    def new_load(self, gr):
        for node in gr.get_ids():
            gr.set_attr(node, self.frm, self.gr.get_attr(node, "prt"))
        return gr

    def new_node_add(self, graph, to, tp, attrs):
        super().new_node_add(graph, to, tp, attrs)
        graph.set_attr(to, self.frm, -1)

    def stop(self):
        return self.chosen == -1


class PACancerDiGrowth(PADiGrowth):
    def __init__(self, graph, schema, deg, model_param):
        super().__init__(graph, schema, deg, model_param)
        self.frm = "prt"
        self.nws = 0
        self.nws_max = 5000
        self.outs = powerlaw(self.nws_max, 1, ac=0.05)

    def prep(self, graph):
        return graph

    def choice(self, graph, sz, tp="in"):
        if tp == "edge":
            nodes = graph.get_nodes_with("nw", 1)
            if nodes.size == 0:
                return None, None

            probs1 = np.array([graph.count_in_degree(node) for node in nodes]) + self.param[0]
            probs2 = np.array([graph.count_out_degree(node) for node in nodes]) + self.param[1]
            probs1 = probs1 / np.sum(probs1)
            probs2 = probs2 / np.sum(probs2)

            return np.random.choice(nodes, 1, replace=False, p=probs1)[0], \
                   np.random.choice(nodes, 1, replace=False, p=probs2)[0]
        else:
            return super().choice(graph, sz, tp=tp)

    def save(self, gr, to):
        for node in gr.get_ids():
            gr.set_attr(node, "nw-"+self.frm, -1)

        while gr.get_nodes_with("nw-"+self.frm, -1).size != 0:
            for node in gr.get_ids():
                if gr.get_attr(node, "nw-"+self.frm) == -1:
                    if gr.get_attr(node, "nw") == 1:
                        nbrs = np.append(gr.get_in_degrees(node), gr.get_out_degrees(node))
                        nodes = np.unique(gr.get_attributes(self.frm, nbrs))
                        if np.sum(np.isin([0], nodes)) == 1:
                            gr.set_attr(node, "nw-" + self.frm, 0)
                        elif np.sum(np.isin([1], nodes)) == 1:
                            gr.set_attr(node, "nw-" + self.frm, 1)
                        else:
                            gr.set_attr(node, "nw-" + self.frm, 2)
                    else:
                        gr.set_attr(node, "nw-" + self.frm, gr.get_attr(node, self.frm))

        estimate_rank(gr, "one", pers=None)
        nx.write_graphml(gr.get_nx_graph(), to + ".graphml")

    def new_load(self, gr):
        for node in gr.get_ids():
            gr.set_attr(node, self.frm, self.gr.get_attr(node, "prt"))
            gr.set_attr(node, "nw", 0)
        return gr

    def new_node_add(self, graph, to, tp, attrs):
        super().new_node_add(graph, to, tp, attrs)
        graph.set_attr(to, self.frm, -1)
        graph.set_attr(to, "nw", 1)

        for node in np.random.choice(graph.get_ids(stable=True), min(graph.size(), int(self.outs[self.nws])), replace=False):
            graph.add_edge(to, node)
            graph.set_edge_data(to, node, attrs[0], attrs[1] + 1)
        self.nws += 1

    def stop(self):
        return self.nws == self.nws_max

    def new_edge_add(self, gr, attrs):
        trg = gr.get_nodes_with("nw", 1)
        if trg.size == 0:
            return

        node1, node2 = self.choice(gr, None, tp="edge")
        if node1 is None:
            return
        gr.add_edge(node1, node2)
        gr.set_edge_data(node1, node2, attrs[0], attrs[1] + 1)



import numpy as np

from dtg.stationary.dierckx import stat_exponential
from dtg.stationary.giratis import stat
from pge.ranks.extrem_onl import NodeExInfo


class WayEx(NodeExInfo):
    @staticmethod
    def get_exes_comm(gr, nodes, params, mn=True):
        res = []
        for params_ in params:
            cls = gr.get_attributes(params_[0], nodes)
            ex_tmp = 1
            ex = 1
            u = np.max(cls)
            for u in np.unique(cls)[::-1]:
                ts = np.array(gr.get_all_short_pathes(nodes[cls > u])[1])
                if ts.size == 0:
                    continue

                if np.max(ts) > 2:
                    ex = min(
                        [
                            1,
                            2
                            * np.mean(ts - 1) ** 2
                            / (np.mean(np.multiply(ts - 1, ts - 2))),
                        ]
                    )
                else:
                    ex = min([1, 2 * np.mean(ts) ** 2 / (np.mean(ts ** 2))])
                if mn:
                    if ex < 1:
                        break
                else:
                    if 1 > ex >= ex_tmp > 0:
                        ex = ex_tmp
                        break
                ex_tmp = ex
            res.append((ex, u))
        return res

    @staticmethod
    def get_gen_exes_comm(gr, nodes, params):
        res = []

        ds = np.array([gr.count_in_degree(node) for node in gr.get_ids()])
        for params_ in params:
            fl_cls = gr.get_attributes(params_[0])
            cls = gr.get_attributes(params_[0], nodes)
            ex_tmp = 1
            ex = 1
            u = np.min(cls)
            for u in np.unique(cls)[::-1]:
                ts = np.array(gr.get_all_short_pathes(nodes[cls > u])[1])
                if ts.size == 0:
                    continue

                t1 = np.mean(ts)
                t2 = np.mean(ts ** 2)
                t3 = np.mean(ts ** 3)
                ex = max(
                    [
                        0,
                        min(
                            [
                                1,
                                np.log(1 - 3 * (t2 - t1) / (t3 - t1))
                                / (np.log(np.sum(ds[fl_cls <= u]) / np.sum(ds))),
                            ]
                        ),
                    ]
                )

                if 1 > ex >= ex_tmp > 0:
                    break
                ex_tmp = ex
            res.append((ex_tmp, u))
        return res

    @staticmethod
    def get_view(gr, nodes, params):
        res = []

        # ds = np.array([gr.count_in_degree(node) for node in gr.get_ids()])
        ds = np.array([gr.count_in_degree(node) for node in nodes])
        for params_ in params:
            cls = gr.get_attributes(params_[0], nodes)

            for u in np.unique(cls)[::-1]:
                ts = np.array(gr.get_all_short_pathes(nodes[cls > u])[1])

                res.append(np.mean(ds[cls > u]) / ts.size)
        return res, np.unique(cls)[::-1]

    @staticmethod
    def get_test_comm(gr, nodes, level, param):
        if nodes.size == 0:
            return (0, 0), (0, 0)

        cls = gr.get_attributes(param, nodes)
        ts = gr.get_all_short_pathes(nodes[cls > level], plain=True)[1]
        if ts.size == 0:
            return (0, 0), (stat(cls), np.mean(cls))

        q = np.sum(gr.get_attributes(param) > level) / gr.size()
        return (
            (
                stat_exponential(ts, q),
                q
                * np.sum([np.sum(ti) for ti in ts])
                / np.sum([np.size(ti) for ti in ts]),
            ),
            (stat(cls), np.mean(cls)),
        )



import numpy as np
import networkx as nx
from itertools import combinations, product


def graph_coherence(graph, nodes):
    L = nx.laplacian_matrix(graph.get_nx_graph()).toarray()
    Q = np.delete(np.delete(L, nodes, axis=0), nodes, axis=1)
    return np.trace(Q)/2


def min_error(graph, k):
    combs = list(combinations(np.arange(graph.size()), k))
    HS = [graph_coherence(graph, comb) for comb in combs]

    res = np.argmin(HS)
    return [list(graph.get_ids())[node] for node in combs[res]]


def kswap(graph, nodes):
    T = np.array(graph.get_ids())
    T = list(T[~np.isin(T, nodes)])

    decreased = True
    while decreased:
        prev = graph_coherence(graph, nodes)
        brk = True
        for i, j in product(nodes, T):
            cp = nodes.copy()
            cp.remove(i)
            if prev > graph_coherence(graph, cp+[j]):
                nodes = cp+[j]
                T.remove(j)
                T.append(i)
                brk = False
                break
        if brk:
            decreased = False

    return nodes


def kgreedy(graph, k):
    res = []
    prev = None

    nodes = list(graph.get_ids())
    while len(res) < k:
        er = [graph_coherence(graph, res+[node]) for node in nodes]
        if prev is None or prev > np.min(er):
            ind = np.argmin(er)
            if hasattr(ind, "__iter__"):
                nd = nodes[ind[0]]
            else:
                nd = nodes[ind]
            res.append(nd)
            prev = np.min(er)
            nodes.remove(nd)
        else:
            break
    return res



import numpy as np
import networkx as nx


from pge.model.growth.basic import BasicGrowth


class FixGrowth(BasicGrowth):
    def __init__(self, graph, schema, deg, model_param, del_typ):
        super().__init__(graph, schema, deg, model_param)
        self.typ = del_typ

    def proceed(self, n, save, attr="cnt"):
        nw_graph = self.gr.clean_copy()
        for node in nw_graph.get_ids():
            nw_graph.set_attr(node, attr, 0)
            nw_graph.set_attr(node, attr + "_end", -1)

        for edge in nw_graph.get_edges():
            nw_graph.set_edge_data(edge[0], edge[1], attr, 0)
            nw_graph.set_edge_data(edge[0], edge[1], attr + "_end", -1)
        graph = self.prep(self.gr.clean_copy())

        count = self.gr.size()
        for _ in np.arange(n):
            new_node = np.random.choice(len(self.schema), p=self.schema)
            if new_node != 1:
                if self.typ == "un":
                    del_node = np.random.choice(graph.get_ids(stable=True))
                else:
                    ids = graph.get_ids(stable=True)
                    vals = graph.get_attributes(attr)
                    del_node = np.random.choice(ids[vals == np.min(vals)])
                nw_graph.set_attr(del_node, attr + "_end", _ + 1)
                for deg_node in nw_graph.get_in_degrees(del_node):
                    nw_graph.set_edge_data(deg_node, del_node, attr + "_end", _ + 1)
                graph.del_node(del_node)

                for node in self.clean(graph):
                    nw_graph.set_attr(node, attr + "_end", _ + 1)
                    for deg_node in nw_graph.get_in_degrees(node):
                        nw_graph.set_edge_data(deg_node, node, attr + "_end", _ + 1)
                    graph.del_node(node)

                self.new_node_add((graph, nw_graph), str(count), new_node, (attr, _))
                count += 1
            else:
                self.new_edge_add((graph, nw_graph), (attr, _))

        for node in graph.get_ids():
            nw_graph.set_attr(node, attr + "_end", _ + 1)
        for edge in graph.get_edges():
            nw_graph.set_edge_data(edge[0], edge[1], attr + "_end", _ + 1)
        nx.write_graphml(nw_graph.get_nx_graph(), save + ".graphml")

    def clean(self, graph):
        return graph


class FixDiGrowth(FixGrowth):
    def new_node_add(self, graph, to, tp, attrs):
        if tp == 0:
            if self.deg[0][0] == "const":
                nodes = self.choice(graph[0], self.deg[0][1], tp="in")
            else:
                nodes = self.choice(graph[0], self.deg[0][0](self.deg[0][1]), tp="in")

            for node in nodes:
                graph[0].add_edge(to, node)
                graph[1].add_edge(to, node)
                graph[1].set_edge_data(node, to, attrs[0], attrs[1] + 1)
        else:
            if self.deg[1][0] == "const":
                nodes = self.choice(graph, self.deg[1][1], tp="out")
            else:
                nodes = self.choice(graph, self.deg[1][0](self.deg[1][1]), tp="out")

            for node in nodes:
                graph[0].add_edge(to, node)
                graph[1].add_edge(to, node)
                graph[1].set_edge_data(node, to, attrs[0], attrs[1] + 1)
        graph[1].set_attr(to, attrs[0], attrs[1] + 1)

    def new_edge_add(self, gr, attrs):
        node1, node2 = self.choice(gr[0], 1, tp="out"), self.choice(gr[0], 1, tp="in")
        gr[0].add_edge(node1, node2)
        gr[1].add_edge(node1, node2)
        gr[1].set_edge_data(node1, node2, attrs[0], attrs[1] + 1)

    def choice(self, gr, sz, tp="in"):
        return []


class FixUnGrowth(FixGrowth):
    def new_node_add(self, graph, to, tp, attrs):
        if self.deg[0] == "const":
            nodes = self.choice(graph[0], self.deg[1])
        else:
            nodes = self.choice(graph[0], self.deg[0](self.deg[1]))

        for node in nodes:
            graph[0].add_edge(to, node)
            graph[1].add_edge(to, node)
            graph[1].set_edge_data(to, node, attrs[0], attrs[1] + 1)
        graph[1].set_attr(to, attrs[0], attrs[1] + 1)

    def new_edge_add(self, gr, attrs):
        nodes = self.choice(gr[0], 2)
        gr[0].add_edge(nodes[0], nodes[1])
        gr[1].add_edge(nodes[0], nodes[1])
        gr[1].set_edge_data(nodes[0], nodes[1], attrs[0], attrs[1] + 1)



import numpy as np
import networkx as nx


def phi(gr, nodes, other=None):
    if gr.directed():
        ind_i = np.arange(gr.size())[[node in nodes for node in gr.get_ids()]]
        if other is not None:
            ind_j = np.arange(gr.size())[[node in other for node in gr.get_ids()]]
        else:
            ind_j = np.arange(gr.size())[[node not in nodes for node in gr.get_ids()]]
        pi = gr.get_stat_dist()
        c = np.sum(pi[ind_i])
        v = np.sum(pi[ind_j])

        p = gr.get_prob_matrix()
        w = np.multiply(p, pi)

        p1 = np.sum(w[np.ix_(ind_j, ind_i)])
        p2 = np.sum(w[np.ix_(ind_i, ind_j)])
        return np.max([p1 / c, p2 / v])
    else:
        try:
            return nx.algorithms.cuts.conductance(gr.get_nx_graph(), nodes, other)
        except:
            return 0


def conductance_speed(gr):
    if nx.is_connected(gr.get_nx_graph()):
        ls = np.sort(nx.normalized_laplacian_spectrum(gr.get_nx_graph()))[1]
        dg_mx = np.max([gr.count_in_degree(node) for node in gr.get_ids()])
        return np.mean([ls/(2*dg_mx), np.sqrt(ls*2/dg_mx)])
    else:
        return 0

def conductance(gr, start_node=None):
    if start_node is None:
        comp = sorted(
            nx.algorithms.community.greedy_modularity_communities(gr.get_nx_graph()),
            key=len,
            reverse=True,
        )
        return conductance_iter(gr, [(list(com), phi(gr, com)) for com in comp])
    else:
        return conductance_iter(gr, [([start_node], phi(gr, [start_node]))])


def conductance_iter(gr, com):
    fx = []

    while len(com) > 0:
        com_ = []
        for comm in com:
            if comm[1] == 0:
                return 0
            comm_ = []
            tst_ = []

            if len(comm[0]) + 1 <= gr.size() / 2:
                for node_out in gr.get_out_degrees(comm[0], un=True):
                    if node_out in comm[0]:
                        continue

                    try:
                        tst_.append(phi(gr, comm + [node_out]))
                        comm_.append(comm + [node_out])
                    except:
                        continue

            for node in comm[0]:
                nw_comm = comm[0].copy()
                nw_comm.remove(node)
                if len(nw_comm) == 0:
                    continue

                gr_ = gr.subgraph(nw_comm)
                if nx.number_connected_components(gr_.get_nx_graph()) != 1:
                    continue

                try:
                    tst_.append(phi(gr, nw_comm))
                    comm_.append(nw_comm)
                except:
                    continue

            if len(tst_) == 0:
                fx.append(comm)
                continue

            if np.min(tst_) < comm[1]:
                rs = np.min(tst_)
                for i in np.arange(len(tst_)):
                    if rs == tst_[i]:
                        com_.append((comm_[i], rs))
            else:
                fx.append(comm)
        com = com_

    return np.min([fx_[1] for fx_ in fx])


def weak_conductance(gr, attr):  # todo вложение
    nx.set_node_attributes(gr.get_nx_graph(), 0, attr)
    comms = sorted(
        nx.algorithms.community.greedy_modularity_communities(gr.get_nx_graph()),
        key=len,
        reverse=True,
    )

    updatable = True
    while updatable:
        comms_ = []
        updatable = False
        for comm in comms:
            ph = phi(gr, comm)
            if ph == 0:
                comms_.append(comm)
                continue

            vars = []
            vars_comm = []
            for node_out in gr.get_out_degrees(comm, un=True):
                if node_out in comm:
                    continue

                try:
                    vars.append(phi(gr, [node_out] + comm))
                    vars_comm.append([node_out] + comm)
                except:
                    continue

            if len(vars) == 0:
                comms_.append(comm)
                updatable = updatable or False
                continue

            mn = np.min(vars)
            if mn <= ph:
                for i in np.arange(len(vars)):
                    if vars[i] == mn:
                        comms_.append(vars_comm[i])
                        updatable = updatable or True
            else:
                comms_.append(comm)
                updatable = updatable or False
        comms = comms_

    for com in comms:
        if len(com) == 1:
            gr.set_attr(list(com)[0], attr, 0)
            continue
        ph = conductance(gr.subgraph(com))
        for node in com:
            gr.set_attr(node, attr, max(gr.get_attr(node, attr), ph))
